#ifndef __GTK_PAPER_TAPE_COLOR__
#define __GTK_PAPER_TAPE_COLOR__

#include <gtkmm/drawingarea.h>
#include <gtkmm/checkmenuitem.h>
#include <gtkmm/menuitem.h>
#include <gtkmm/sizegroup.h>
#include <gtkmm/actiongroup.h>
#include <gtkmm/toggleaction.h>
#include <gtkmm/colorbutton.h>
#include <gtkmm/action.h>
#include <vector>
#include <iostream>

#include <cairomm/pattern.h>

namespace Gtk
{ class PaperTapeColor; }
#include "gtkpapertape.h"
#include "gtkpapertapeview.h"
#include "lochstreifen.h"

namespace Gtk {


/**
 * The PaperTapeColor class represents a color which will be used
 * to paint some component on the PAPERTAPE. It also implements
 * an Action -- that is: This class is an Action. It serves a
 * nice Menuitem for color selection, as well as an auxillary
 * toggle Menuitem to switch the display of that LOCHSTREIFEN
 * component on or off.
 * This is mainly intended to be used with PaperTapeView. That
 * class automatically initializes such a PaperTapeColor class
 * for every LOCHSTREIFEN cairo_pattern_t object. These Color
 * Objects are stored in a map.
 * @see PaperTapeView::colors
 **/
class PaperTapeColor : public Action {
private:
	/// The toggle action as a "nested action", somewhat quick&dirty
	Glib::RefPtr<ToggleAction> toggle_action;
	/// Much more Quick and Dirty: The menuitem which will be
	/// generated at run time by create_menu_item_vfunc.
	/// Stored here for enable/disable access from toggled()
	MenuItem* menu_item;

	/// The View for which this color is allocated.
	/// Deprecated -- not used at all!
	PaperTapeView* view;

	/// can this element be disabled with
	/// display_toggle?
	bool switchable;
	/// Title for color dialog
	Glib::ustring dialog_title;
	/// Label for menu color item
	Glib::ustring menu_label;
	/// Label for toggle menu item
	Glib::ustring toggle_label;
	/// Action name (base name for toggle and color action)
	Glib::ustring action_name;
	/// Target pattern which will be affected
	cairo_pattern_t **target;

	/// The current value of the color
	cairo_pattern_t *value;

	/// The constructor is private, too. USe construct(...)
	/// to get a RefPtr instance.
	PaperTapeColor(const Glib::ustring& name);

	/// see signal_color_changed() for an explanation.
	sigc::signal<void,PaperTapeColor*> color_changed;
protected:
	/// This method creates the ColorMenuItem thing
	/// at runtime.
	virtual Widget* create_menu_item_vfunc();
	/// This is used to keep the Gtk Action mechanism from
	/// overwriting my own MenuItem with a stupid GtkLabel.
	virtual void connect_proxy_vfunc (Widget* proxy) {
		std::cout << "They'll want to proxy some widget!\n";
	}

	/// This is currently not used (Static initialisation issue --
	/// GLIB not initialized!), but is intended to make the
	/// menu_items over many PaperTapeColors look equal.
	static const Glib::RefPtr<SizeGroup> label_group;

	/// An internal signal for the color_button
	void color_set(); // CB
	/// An internal signal for the toggle_action
	void toggled(); // CB

	/// The color button, quick and dirty stored here for
	/// access e.g. at color_set() or set().
	ColorButton color_button;
	
public:
	/**
	 * The official constuctor, it will return a RefPtr reference,
	 * like all the Gtk::Action classes have.
	 * @param view The View which called this
	 * @param target The address of the LOCHSTREIFEN target pattern link
	 * @param action_base_name The base name, "_color" and "_toggle" will be appended
	 * @param dialog_title The title of the Color selection dialog
	 * @param menu_label The pango markuped label of the ColorMenuItem
	 * @param toggle_label THe label of the ToggleMenuItem. Leave empty if not toggled.
	 * @returns A smart pointer to a PaperTapeColor.
	 **/
	static Glib::RefPtr<PaperTapeColor> create(
		PaperTapeView* view,
		cairo_pattern_t** target,
		const Glib::ustring& action_base_name,
		const Glib::ustring& dialog_title,
		const Glib::ustring& menu_label,
		const Glib::ustring& toggle_label
	);

	/// Well, use this to find out if you need to export the
	/// toggle_action Action. It won't be useful if the element that
	/// this color represents is not intented to be turned off.
	bool is_switchable() { return switchable; }
	
	/// Give out if the element that this color represents is currently
	/// displayed.
	bool is_displayed() { return toggle_action->get_active(); }
	
	/// Get the toggle action. Don't use it if is_switchable()
	/// returns false.
	Glib::RefPtr<ToggleAction> get_toggle_action() { return toggle_action; }
	
	/**
	 * This is a user function (for you!) to set the Color directly.
	 * It will send all signals and update all buttons automatically.
	 * Note: This won't turn on the color, if turned off currently.
	 * Use get_toggle_action()->set_active(true); for that purpose.
	 *
	 * @param red Percentual red, from 0.0 to 1.0
	 * @param green These fields are just the same...
	 * @param blue ...than cairo_pattern_create_from_rgba...
	 * @param alpha ...wants them.
	 **/
	void set(double red, double green, double blue, double alpha);
	/// This makes the same, it just copies your pattern.
	void set(cairo_pattern_t* pattern);
	
	/**
	 * This signal is emitted whenever the color is changed (by used input or
	 * exiplit set() call) or the visibility of the element which is controlled
	 * by this Color object has been toggled.
	 * Therefore you can use this signal quite well to perform exposure events,
	 * etc.
	 * You'll recieve a pointer to this object with this signal. Don't free it.
	 **/
	sigc::signal<void,PaperTapeColor*> signal_color_changed() { return color_changed; }
	
	/// Don't know if anybody needs this function.
	MenuItem* create_menu_item() {
		return (MenuItem*)create_menu_item_vfunc();
	}
};


} // namespace
#endif // TAPE_COLOR
