#ifndef __PAPER_TAPE_FILE_H__
#define __PAPER_TAPE_FILE_H__

/* $Id$ */

/* visualisation/papertape/file.h
 *
 * Copyright (C) 2008 Sven Koeppel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string>

namespace PaperTape
{ class File; }
#include "../lochstreifen.h"
#include "view.h"

namespace PaperTape {

/**
 * The File class hold the contents of a file (always completely in RAM).
 * This is the "model" class in the PaerTape concept. This class a nice
 * interface with many handy methods, even GTK dialogs for opening/closing/....
 *
 * The class doesn't hold instances of View or Chrome. For proper dialog
 * working, you are encouraged to use set_parent_window(Gtk::Window) or
 * the appropriate constructor.
 *
 * This class uses sigc++ signals that are fired whenever the internal binary
 * data array is changed.
 *
 * Compared to the Gtk+ text widget system, this class is comparable to the
 * GtkTextBuffer. Therefore there's the nested iterator, working like GtkTextIter.
 *
 **/
class File {
	PaperTape *controller;

	byte_t *data;
	int data_len;
	
public:
	/**
	 * This signal is emitted whenever the data are changed.
	 **/
	sigc::signal<void, byte_t*, int> data_changed;

	PaperTapeFile(PaperTape* controller);
	
	// the nested iterator class
	class iterator;
	friend class iterator;
	
	/**
	 * This iterator shall feel like a row_t, which is an int typedef,
	 * so this iterator shall feel like an int. There's not much overhead
	 * when creating it on run time with the File utility methods.
	 * It acts and feels like any ordinary C++ iterator.
	 *
	 **/
	class iterator {
		bool backward_byte();
		bool backward_bytes(int count);
		bool backward_search(byte_t value);
		bool backward_search(byte_t value, const iterator& limit);
		bool backward_visible_byte(); // viewport (scrolling!)
		int compare(const iterator& rhs) const;
		// like operator>: returns negative if lhs is less than rhs, positive if lhs is greater than rhs, and 0 if they're equal.
		
		bool forward_byte();
		bool forward_bytes(int count);
		bool forward_search(byte_t value);
		bool forward_visible_byte();
		bool forward_to_end(); // to end of file

		// getters
		File& get_file() const; // returns the associated file
		byte_t get_byte() const; // returns the byte which the iterator is on.
		int get_pos() const; // returns position from 0 in the file
		int get_visible_pos() const; // returns position from viewport
		
		bool in_range(const TextIter& start, const TextIter& end) const; // wheter iter falls in range [start, end[
		bool is_cursor_position() const;
		bool is_end() const;
		bool is_start();	

		iterator& operator++() { forward_char(); return *this;  }
		iterator& operator--() { backward_char(); return *this; }

		// auch intelligent:
		// void order(iterator& second) // swaps the value of first and second if second comes before first in file

		// setters
		void set_byte(row_t row_number); // if row_number > file size => at end of file
	}

	// hier auch Funktionen zum LABEL
	// generieren, Daten abndern, etc.

	// File changing and getting
	/// get a copy of the byte at position "pos".
	inline byte_t get_byte(const row_t& pos);
	/// get the value of the bit at that position
	inline bool get_bit(const row_t& pos, const track_t& track);	
	/// change byte at position "pos" to "value"
	inline bool change_byte(const row_t& pos, byte_t value);
	/// change the bit at position "pos", bit "track" to "value"
	inline bool change_bit(const row_t& pos, const track_t& track, bool value);
	/// toggle the bit at position "pos", bit "track"
	inline bool toggle_bit(const row_t& pos, const track_t& track);
	//void push_bytes
	//void pop_bytes
	// etc.
	
	// clipboard interaction
	void cut_clipboard(const Glib::RefPtr<Clipboard>& clipboard);
	void copy_clipboard(const Glib::RefPtr<Clipboard>& clipboard);
	void paste_clipboard(const Glib::RefPtr<Clipboard>& clipboard); // at current position
	void paste_clipboard(const Glib::RefPtr<Clipboard>& clipboard, const iterator& override_location);
	
	// getting iterators
	iterator start();
	iterator end();
	iterator get_cursor();
	
	erase(const iterator& range_begin, const iterator& range_end);
	erase_selection(); // delete current selection
	
	bool has_selection();
	

	// interne Low Level Funktionen.
	// (intern = Gtk::PaperTape)
	void set(byte_t* data, int data_len);
	byte_t* get(iterator& start, iterator& end); // get a range. this WON'T COPY data, just data[start].
	
	void add_null_bytes(int start, int end);
	
	
	// File saving and opening
	/// GUI function: Show dialog for file opening
	void open();
	
	/// Whether the file was modified since the last set_modified call
	bool get_modified();
	/// "Brute force" change the modification state. Not supposed to be used
	/// (save() calls it itself).
	void set_modified(bool setting);
	
	/// open that file (non GUI)
	/// @throws GError if an error occurred
	void open(const std::string& filename);
	
	/// if filename omitted, save under same name.
	bool save(const std::string& filename = std::string());
	
	/// GUI function: SaveAs dialog incl. handling
	void saveAs();
	
	/// GUI function: Close Model (with dialog)
	/// and destroy Object (notify controller/view).
	/// if you want no GUI, use the destructor
	/// directly (obj.~...Model())
	void close();
	
	/// GUI function: Reload file. Same as
	/// close(); new Model(same_file_name);
	void reload();

};

// inline functions:
byte_t PaperTapeFile::get_byte(const row_t& pos) {
	if(pos > 0 && pos <= data_len) {
		return data[pos];
	} else
		return 0; // muss ich mir noch ueberlegen
}

bool PaperTapeFile::get_bit(const row_t& pos, const track_t& track) {
	return (get_byte(pos) >> track) & 0x01 == 0x01;
}

bool PaperTapeFile::change_byte(const row_t& pos, byte_t value) {
	if(pos > 0 && pos <= data_len) {
		data[pos] = value;
		data_changed.emit(data, data_len);
		return true;
	} else
		return false;
}

bool PaperTapeFile::change_bit(const row_t& pos, const track_t& track, bool value) {
	if(pos > 0 && pos <= data_len && track != LOCHSTREIFEN_NO_TRACK) {
		data[pos] = value
			? data[pos] | (0x01 << track)           // turn on
			: data[pos] & (0xFF ^ (0x01 << track)); // turn off
		data_changed.emit(data, data_len);
		return true;
	} else
		return false;
}

bool PaperTapeFile::toggle_bit(const row_t& pos, const track_t& track) {
	if(pos > 0 && pos <= data_len && track != LOCHSTREIFEN_NO_TRACK) {
		data[pos] ^= (0x01 << track); // XOR toggle.
		data_changed.emit(data, data_len);
		return true;
	} else
		return false;
}

}; // Namespace Gtk
#endif // __PAPER_TAPE_FILE_H
