#ifndef __PAPER_TAPE_ZOOM__
#define __PAPER_TAPE_ZOOM__

/* $Id$ */

/* visualisation/papertape/zoom.h
 *
 * Copyright (C) 2008 Sven Koeppel
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gtkmm/dialog.h>
#include <gtkmm/actiongroup.h>
#include <gtkmm/action.h>
#include <gtkmm/spinbutton.h>
#include <map>

// uses a bit of old style GTK+
#include <gtk/gtk.h>

namespace PaperTape
{ class Zoom; }
#include "view.h"

namespace PaperTape {

enum Rotate {
	ROTATE_90_CW,
	ROTATE_90_CCW,
	NO_ROTATION
}

enum Flip {
	FLIP_HOR,
	FLIP_VER,
	NO_FLIP
};

enum AutoZoomType {
	ZOOM_LOCHSTREIFEN_WIDTH,
	ZOOM_LOCHSTREIFEN_LENGTH
};

/**
 * The Zoom class is an extension to the Cairo Translation Matrix.
 * It features a nice Dialog for direct matrix editing, sigc++ signals,
 * automatic matrix generation from presets (Flipping, Rotation, Offsets)
 * and a nice clean object oriented interface.
 *
 * There's no real relation to PaperTape::View or any other class, so
 * you could use this class simply on your own.
 **/

class Zoom : public Cairo::Matrix {
	/**
	 * The default matrix. This is the return point whenever
	 * you think your matrix is bad, and you want to reset() it.
	 * Change it with set_default() or via the constructor arguments.
	 * Normally this should be something like 1:1 "original" view
	 *
	 **/
	Cairo::Matrix default_matrix;

	/**
	 * For automatical settings
	 *
	 **/
	double rotation;
	Flip flip;
	double zoom_factor;

	/**
	 * Switch: Manual settings (raw matrix editing) or
	 * automatical matrix generation.
	 **/
	bool manually;
	
	void update_dialog_on_zoom_changes();
	void on_toggle_manually(CheckButton* manually_toggle);
	void on_change_field(double* field);
	void calc_matrix();

	Dialog* matrix_dialog;
	std::map<double*, SpinButton*> matrix_entries;
	
	// auto zoom:
	bool auto_zoom_enabled;
	sigc::connection auto_zoom_resize_con;
	bool on_auto_zoom_resize(GdkEventConfigure* event);
	
	static GtkActionEntry ui_action_entries[];

	/// When the zoom has changed somehow, this is emitted (and recieved ;-))
	sigc::signal<void> zoom_changed;
public:
	PaperTapeZoom(double default_rotation = 0.0, double default_zoom_factor = 1.0, Flip default_flip = NO_FLIP);

	void set_default(double rotation = 0.0, double zoom_factor = 1.0, Flip flip = NO_FLIP);

	/**
	 * Resets the matrix to the default. This keeps the manual control,
	 * doen't set it back to automatic control.
	 **/
	void reset();

	void set_rotate(double radians, bool relative);
	void set_zoom(double zoom_factor, bool relative);
	void set_flip(PaperTapeFlip flip, bool relative);
	
	void set_auto_zoom(PaperTapeView& view, PaperTapeAutoZoomType type);
	void disable_auto_zoom();
	
	bool is_controlled_manually() { return manually; }
	void show_dialog();
	
	Glib::RefPtr<ActionGroup> get_action_group();
	void transform_action(const Glib::RefPtr<Action>& action);

	sigc::signal<void> signal_changed() {
		return zoom_changed;
	}

	PaperTapeZoom& operator=(const Cairo::Matrix& right) {
		// self-assignment makes not much sense ;-)
		if(this == &right) return *this;
		this->xx = right.xx; this->xy = right.xy; this->x0 = right.x0;
		this->yx = right.yx; this->yy = right.yy; this->y0 = right.y0;
		return *this;
	}

}; // class

}; // namespace PaperTape

#endif // __PAPER_TAPE_ZOOM__
